/**
 * Finds an eligible loop from the given boolean road grid
 * Always returns an array - may be empty if no path found
 */
var RACE_PATH_MIN_LENGTH = 40;
var RACE_PATH_VERY_MIN_LENGTH = 5;
// returns possible starting positions
var findStartPoints = function (booleanGrid, townHallIndex) {
    // Look for all eligible road tiles around town hall, prefer the tile with lowest distance from base tile
    var baseRow = townHallIndex[0];
    var baseCol = townHallIndex[1];
    var gridSize = booleanGrid.length;
    var foundPoints = {};
    var eligibleIndices = [];
    var tlR = Math.max(baseRow - 2, 0);
    var tlC = Math.max(baseCol - 2, 0);
    var trC = Math.min(gridSize - 1, baseCol + 1);
    var blR = Math.min(gridSize - 1, baseRow + 1);
    var checkAndAdd = function (r, c) {
        var k = r + "_" + c;
        if (booleanGrid[r][c] && !foundPoints[k]) {
            foundPoints[k] = 1;
            eligibleIndices.push([r, c]);
        }
    };
    var col, row;
    // left to right
    for (col = tlC; col <= trC; col++) {
        checkAndAdd(tlR, col);
        if (tlR !== blR) {
            checkAndAdd(blR, col);
        }
    }
    // top to bottom
    if (tlR + 1 <= blR - 1) {
        for (row = tlR + 1; row <= blR - 1; row++) {
            checkAndAdd(row, tlC);
            if (tlC !== trC) {
                checkAndAdd(row, trC);
            }
        }
    }
    var preferred = [baseRow + 1, baseCol];
    if (eligibleIndices.filter(function (i) {
        return i[0] === preferred[0] &&
            i[1] === preferred[1];
    }).length) {
        eligibleIndices.unshift(preferred);
    }
    return eligibleIndices;
};
// track marked roads + spawns ever
var globalMarked = [];
// main loop finder
var findRacePath = function (booleanGrid, townHallPoint) {
    // initialize global mark tracker
    for (var i = 0; i < booleanGrid.length; i++) {
        globalMarked[i] = [];
    }
    // flood fill from starting point until loop found!
    var possibleStarts = findStartPoints(booleanGrid, townHallPoint);
    while (possibleStarts.length) {
        var start = possibleStarts.shift();
        if (!globalMarked[start[0]][start[1]]) {
            var path = findRacePathFromStart(booleanGrid, start);
            if (path.length)
                return path;
        }
    }
    return [];
};
var findRacePathFromStart = function (booleanGrid, startRoadTile) {
    var paths = [];
    var marked = [];
    for (var i = 0; i < booleanGrid.length; i++) {
        marked[i] = [];
    }
    // floodfill from startRoadTile
    var queue = [[startRoadTile[0], startRoadTile[1], 0]];
    marked[startRoadTile[0]][startRoadTile[1]] = {
        index: [startRoadTile[0], startRoadTile[1]],
        prev: null
    };
    var lastLink = null;
    var _loop_1 = function () {
        var index = queue.shift();
        var depth = index[2];
        mapTouchingTiles(index[0], index[1], booleanGrid.length, function (r, c) {
            if (!booleanGrid[r][c]) {
                return;
            }
            globalMarked[r][c] = 1; // for starting points
            if (marked[r][c]) {
                return;
            }
            var link = {
                index: [r, c],
                prev: marked[index[0]][index[1]]
            };
            lastLink = link;
            marked[r][c] = link;
            if (depth >= RACE_PATH_MIN_LENGTH) {
                paths.push(link);
                return;
            }
            queue.push([r, c, depth + 1]);
        });
    };
    while (queue.length) {
        _loop_1();
    }
    if (!paths.length) {
        // no suitable length path found, just use closest
        paths = [lastLink];
    }
    // convert loops from linked list to array
    var loopsConverted = [];
    paths.forEach(function (node) {
        var fullPath = [];
        while (node) {
            fullPath.unshift([node.index[0], node.index[1]]);
            node = node.prev;
        }
        if (fullPath.length > RACE_PATH_VERY_MIN_LENGTH) {
            loopsConverted.push(fullPath);
        }
    });
    // todo: decide which path to use (best fit for min path and min laps)
    if (loopsConverted.length) {
        return randomChoice(loopsConverted);
    }
    else {
        return [];
    }
};
function randomChoice(arr) {
    if (arr.length === 0) {
        return null;
    }
    return arr[randomInRange(0, arr.length - 1)];
}
function randomInRange(min, max) {
    return Math.min(min + Math.floor(Math.random() * (max - min + 1)), max);
}
function mapTouchingTiles(baseRow, baseCol, size, fn, allowInvalidIndex) {
    if (allowInvalidIndex === void 0) { allowInvalidIndex = false; }
    [[0, -1], [0, 1], [1, 0], [-1, 0]].map(function (d) {
        var newRow = baseRow + d[0];
        var newCol = baseCol + d[1];
        if ((newRow >= 0 && newRow < size && newCol >= 0 && newCol < size) ||
            allowInvalidIndex) {
            fn(newRow, newCol);
        }
    });
}
onmessage = function (e) {
    var obj = JSON.parse(e.data);
    postMessage(JSON.stringify(findRacePath(obj.booleanGrid, obj.startPoint)));
};
